/***************************************************************************
 *
 * SinglyLinkedList2.java
 *
 * More advanced implementation of the List interface using a singly
 * linked list with references to both the head and tail Node's in the list
 *
 *****************************************************************************/

import java.util.*;

public class SinglyLinkedList2<E> implements List<E> {
// Remark: All Implemented Interfaces:
//java.lang.Iterable<E>, java.util.Collection<E>, java.util.List<E> 

 /*******************************************************
 *
 *  The Node class
 *
 ********************************************************/
   protected static class Node<E> {
      private E data;
      private Node<E> next;

      public Node(E data, Node<E> next) {
         this.data = data;
         this.next = next;
      }
   }  //end-class-Node


   private Node<E> head;
   private Node<E> tail;
   private int size = 0;


 /**
   *  Constructs an empty list
   */
   public SinglyLinkedList2()
   {
      head = tail = null;
   }

 /**
   *  Returns true if the list is empty
   *
   */
   public boolean isEmpty()
   {
      return head == null;
   }

 /**
   *  Returns the length of the list
   *
   */
   public int size()  
   {
      return size;
   }

 /**
   *  Inserts a new node at the beginning of this list.
   *
   */
   public void addFirst(E item)
   {
      head = new Node<E>(item, head);
      if (tail == null) tail = head; // adds the first node
      size++;	
   }

 /**
   *  Returns the first element in the list.
   *
   */
   public E getFirst()
   {
      if(head == null) throw new NoSuchElementException();

      return head.data;
   }

 /**
   *  Removes the first element in the list.
   *
   */
   public E removeFirst()
   {
      E tmp = getFirst();
      head = head.next;
      size--;
      if (size == 0) tail = null;
      return tmp;
   }

 /**
   *  Inserts a new node to the end of this list.
   *
   */
   public void addLast(E item)
   {
      if( head == null)
         addFirst(item);
      else {
    	 tail.next = new Node<E>(item, null);
    	 tail = tail.next;
    	 size++;
      }
   }


 /**
   *  Returns the last element in the list.
   *
   */
   public E getLast()
   {
      if(head == null) throw new NoSuchElementException();
      return tail.data;
   }


 /**
   *  Removes the last element in the list.
   *
   */
   public E removeLast()
   {  

	Node<E> prev;
	E elem; 

    if(head == null) throw new NoSuchElementException();

    if (head.next == null) return removeFirst();
	// there are at least two nodes
    prev = head;	
    while(prev.next != tail) prev = prev.next;
    elem = tail.data;	
    prev.next = null;
    tail = prev;
    size--;
    return elem;
   }


 /**
   *  Removes all nodes from the list.
   *
   */
   public void clear()
   {
      head = tail = null; size=0;
   }

/**
   *  Inserts a new node to the end of this list.
   *
   */
   public boolean add(E item)
   {
      this.addLast(item);
      return true;	
   }

/**
   *  Inserts the specified item at the specified position in this list.
   *
   */
   public void add(int index, E item)
   {
      if(index == 0) 
      {
         addFirst(item); 
         return;
      }

      Node<E> cur = head;
      int count = 0;	

      while(cur != null && count < index-1) {
	cur = cur.next;
	count++;
      }

      if(cur != null) {
         cur.next = new Node<E>(item, cur.next);
         size++;
         // if inserts at the last position
         if (cur == tail) tail = tail.next;
      } else throw new IndexOutOfBoundsException();
   }


 /**
   *  Returns true if this list contains the specified element.
   *
   */
   public boolean contains(Object x)
   {
      for(E tmp : this)
         if(tmp.equals(x)) return true;

      return false;
   }


 /**
   *  Returns the data at the specified position in the list.
   *
   */
   public E get(int pos)
   {
      if (head == null) throw new IndexOutOfBoundsException();

      Node<E> tmp = head;
      for (int k = 0; k < pos; k++) {
	tmp = tmp.next;
     	if( tmp == null) throw new IndexOutOfBoundsException();
      }
      return tmp.data;
   }


 /**
   *  Returns a string representation
   *
   */
   public String toString()
   {
      StringBuffer result = new StringBuffer();  
      for(Object x : this)
      	result.append(x + "; ");

      return result.toString();
   }

 /**
   *  Inserts a new node after a node containing the key.
   *
   */
   public void insertAfter(E key, E toInsert)
   {
      Node<E> cur = head;

      while(cur != null && !cur.data.equals(key)) cur = cur.next;

      if(cur != null) {
    	  cur.next = new Node<E>(toInsert, cur.next);
    	  if (cur == tail) tail = tail.next;
    	  size++;
      }
   }

 /**
   *  Inserts a new node before a node containing the key.
   *
   */
   public void insertBefore(E key, E toInsert)
   {
      if(head == null) return;

      if(head.data.equals(key))
      {
         addFirst(toInsert); 
         return;
      }

      Node<E> prev = null;
      Node<E> cur = head;

      while(cur != null && !cur.data.equals(key))
      {
         prev = cur;
         cur = cur.next;
      }
      //insert between cur and prev
      if(cur != null) {
         prev.next = new Node<E>(toInsert, cur);
	 size++;
      }
   }

 /**
   *  Removes the first occurrence of the specified element in this list.
   *
   */
   public boolean remove(Object key)
   {
      if(head == null)
         // throw new RuntimeException("cannot delete");
	return false;

      if( head.data.equals(key) ) {
    	  removeFirst();
    	  return true;
      }

      Node<E> prev = null;
      Node<E> cur  = head;


      while(cur != null && !cur.data.equals(key) )
      {
         prev = cur;
         cur = cur.next;
      }

      if(cur == null)
         // throw new RuntimeException("cannot delete");
    	 return false;

      //delete cur node
      prev.next = cur.next;
      if (cur == tail) tail = prev;
      size--;
      return true;	
   }

 /*******************************************************
 *
 *  The Iterator class
 *
 ********************************************************/


   public Iterator<E> iterator()
   {
      return new LinkedListIterator();
   }


   private class LinkedListIterator  implements Iterator<E>
   {
      private Node<E> nextNode;

      public LinkedListIterator()
      {
         nextNode = head;
      }

      public boolean hasNext()
      {
         return nextNode != null;
      }

      public E next()
      {
         if (!hasNext()) throw new NoSuchElementException();
         E res = nextNode.data;
         nextNode = nextNode.next;
         return res;
      }

      public void remove() { throw new UnsupportedOperationException(); }

   }  //end class LinkedListIterator


    /************* HOMEWORK *************************
    public E remove(int index);
    public E set(int index, E element);
    public boolean equals(Object obj);
    public SinglyLinkedList2<E> copy();
	public SinglyLinkedList2<E> reverse();
	public void reverseThisList();
	public Object[] toArray();
    *************************************************/

     /**
       *  Removes the element at the specified position in this list
       */   

         public E remove (int index) {
            if(head == null) throw new RuntimeException("cannot delete");
            if (index == 0) return this.removeFirst();	
            Node<E> prev = head;
            int pos = 1; 	
            while (prev.next != null && pos++ != index) prev = prev.next;
            if (prev.next == null) throw new RuntimeException("cannot delete");
            else {		
               //delete next node
      	   E obj = prev.next.data;
                 prev.next = prev.next.next;
                 if (prev.next == null) tail = prev;
                 size--;
                 return obj;
            }
         }

     /**
       *  Replaces the element at the specified position in this list
       *  with the specified element.
       *  Returns the element previously at the specified position
       */
       public E set(int index, E element)
       {
    	   Node<E> cur = head;
    	   int count = 0;
    	   while(cur != null && count < index) {
    		   cur = cur.next;
    		   count++;
    	   }
    	   if (cur != null) {
    		   E oldElement = cur.data;
    		   cur.data = element;
    		   return oldElement;
    	   }
    	   else throw new IndexOutOfBoundsException();

	   }
       
       /**
        * Two SinglyLinkedList2 instances are said to be equal if both
        * the instances contain the same elements in exactly the same order.
        * */
          public boolean equals(Object obj) {

          if (this == obj) return true;
          if (obj == null) return false;
          if (!(obj instanceof SinglyLinkedList2<?>) ) return false;
          SinglyLinkedList2<?> l2 = (SinglyLinkedList2<?>) obj;

          Node<?> tmp1 = this.head,
       		   tmp2 = l2.head;
          while(tmp1!= null && tmp2!= null){
       	//Confronto dei corrispondenti elementi	
       	if(! tmp1.data.equals(tmp2.data)) return false;
       	else {
       	      tmp1 = tmp1.next;
                     tmp2 = tmp2.next;
              	}
          }
          // verifica delle lunghezze delle liste 
          if (tmp1 == null && tmp2 == null) return true;
          else return false;

       }

       /**** VERSIONE RICORSIVA ****/

       public boolean equalsRec(Object obj) {

          if (this == obj) return true;
          if (obj == null) return false;
          if (!(obj instanceof SinglyLinkedList2<?>) ) return false;
          return equalsRec(this.head, ((SinglyLinkedList2<?>)obj).head);
       }


       private boolean equalsRec(Node<E> n1, Node<?> n2) {
          // condizioni di terminazione
          if (n1 == null && n2 == null) return true; 
          else if (n1 == null || n2 == null) return false;
          // passo ricorsivo
          if ( ! ((n1.data).equals(n2.data)) )
       	return false;
          else return equalsRec(n1.next, n2.next);

       }

          /**
           *  Reverses the list without deleting the initial list
           *  Complexity: O(n)
           */
           public SinglyLinkedList2<E> reverse()
           {
              
              SinglyLinkedList2<E> list = new SinglyLinkedList2<E>();
              Node<E> cur = head;
              while(cur != null)
              {
                 list.addFirst( cur.data );
                 cur = cur.next;
              }
              return list;
           }
           
           
           /**
            *  Reverses this list
            *  Complexity: O(n)
            */
            public void reverseThisList()
            {        
               SinglyLinkedList2<E> list = new SinglyLinkedList2<E>();
               while(!isEmpty()) list.addFirst(this.removeFirst());
               head=list.head;
               tail=list.tail;
               size=list.size();
            }



           /**
            *  Reverses the list without deleting the initial list
            *  Complexity: O(n)
            */
            public SinglyLinkedList2<E> reverse2()
            {
               SinglyLinkedList2<E> list = new SinglyLinkedList2<E>();
               Iterator<E> itr = this.iterator();
               while (itr.hasNext()) list.addFirst(itr.next());
               return list;
            }


         /**
           *  Returns a copy of the immutable list
           *  Complexity: O(n)
           */
           public  SinglyLinkedList2<E> copy1()
           {
              SinglyLinkedList2<E> twin = new SinglyLinkedList2<E>();
              Node<E> tmp = head;
              while(tmp != null)
              {
                 twin.addLast( tmp.data );// O(1) in questa implementazione
                 tmp = tmp.next;
              }

              return twin;
           }

         /**
           *  Returns a copy of the immutable list
           *  Complexity: O(n)
           */
           public SinglyLinkedList2<E> copy2()
           {
              SinglyLinkedList2<E> twin = new SinglyLinkedList2<E>();
              Node<E> tmp = head;
              while(tmp != null)
              {
                 twin.addFirst( tmp.data );
                 tmp = tmp.next;
              }

              return twin.reverse();
           }

  /**
   * toArray() method returns an array containing all of
   * the elements in this list in proper sequence (from
   * first to last element). 
   * */ 

   public Object[] toArray() {
         Object[] array;
         int i=0;	
         if (head == null) return null;
         array = new Object[size];
         for (E item: this) {array[i++] = item; }
         return array;

     }
   
   
        /*
        *   Unsupported methods
        */
          public List<E> subList(int fromIndex, int toIndex)
          // Returns a view of the portion of this list between the specified fromIndex, inclusive, and toIndex, exclusive. 
          { throw new UnsupportedOperationException(); }

          public int indexOf(Object o)
          // Returns the index of the first occurrence of the specified element in this list,
          // or -1 if this list does not contain the element.
          { throw new UnsupportedOperationException(); }

          public int lastIndexOf(Object o)
          // Returns the index of the last occurrence of the specified element in this list,
          // or -1 if this list does not contain the element.
          { throw new UnsupportedOperationException(); }

          public ListIterator<E> listIterator()
            	// Returns a list iterator over the elements in this list (in proper sequence)
          { throw new UnsupportedOperationException(); }

          public ListIterator<E> listIterator(int index)
            	// Returns a list iterator over the elements in this list (in proper sequence),
       	// starting at the specified position in the list.
          { throw new UnsupportedOperationException(); }

          public boolean retainAll(Collection<?> c)
          { throw new UnsupportedOperationException(); }

          public boolean removeAll(Collection<?> c)
          { throw new UnsupportedOperationException(); }

          public boolean addAll(Collection<? extends E> c)
          { throw new UnsupportedOperationException(); }

          public boolean addAll(int index, Collection<? extends E> c)
          { throw new UnsupportedOperationException(); }

          public boolean containsAll(Collection<?> c)
          { throw new UnsupportedOperationException(); }

          public <T> T[] toArray(T[] a)
          { throw new UnsupportedOperationException(); }

} // end-class SinglyLInkedList2